"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

textExample.py

This file used to provide an example of most of the high performance usage of
PyGL3Display 2d Sprites in version 1.1 and below. In Version 1.9 and onwards,
it exists to demonstrate SpriteGroups, even though these are under threat of
depreciation.
"""

from exampleskeleton import *
import pygame
    
class SpriteFont(object):
    def __init__(self, font, size):
        """Prerender the font"""
        chars  = ' abcdefghijklmnopqrstuvwxyzQWERTYUIOPASDFGHJKLZXCVBNM,.'
        chars += '/?!"$%^\\\';&*()-_=+1234567890'
        font = pygame.font.Font(pygame.font.match_font(font), size)
        self.fontmap = {}
        self.height = 0
        for char in chars:
            character = font.render(char, True, (255,255,255,255))
            self.height = max((self.height, character.get_height()))
            self.fontmap[char] = (PygameSurface.surfaceify(character), 
                                 character.get_width())
        self.spriteBuffer = SpriteBuffer(100)
    
    def makeString(self, string):
        """Creates a SpriteGroup which renders the string."""
        lines = string.split('\n')
        sprites = []
        ypos = 0
        maxXpos = 0
        for line in lines:
            xpos = 0
            for char in line:
                if char in self.fontmap:
                    if char != ' ':
                        newSprite = self.spriteBuffer.createSprite(self.fontmap[char][0])
                        newSprite.setOffsets(topLeftOffsets, relative=True)
                        newSprite.setPosition((xpos, ypos))
                        sprites.append(newSprite)
                    xpos += self.fontmap[char][1]
            maxXpos = max(xpos, maxXpos)
            ypos += self.height
        ret = SpriteGroup(sprites)
        return ret

class TestApp(ExampleSkeleton):
    def __init__(self):
        super(TestApp, self).__init__('Text example', (800, 800))
        self.string = ('Things you can do: \n' +
                       ' r: cycle red values \n' +
                       ' g: cycle green values \n' +
                       ' b: cycle blue values \n' +
                       ' arrow keys: move text \n' +
                       ' y: clockwise rotate \n' +
                       ' ijkl: increase / decrease x/y scale\n' +
                       'I\'m afraid that holding keys down won\'t work.\n'+
                       'Apologies for the problems with this sample.\n ' + 
                       '      Habilain')
        self.origx, self.origy = (0, 0)
        self.r, self.g, self.b = (255, 255, 255)
        self.posx, self.posy = (0, 0)
        self.scalex, self.scaley = (1, 1)
        self.rotation = 0
        self.z = 0
        self.frames = 0
        self.countedTime = 0
        
    def init(self):
        self.font = SpriteFont('dejavu-sans', 20)
        self.text = self.font.makeString(self.string)
        
    def eventHandler(self, event):
        if event.type == KEYDOWN:
            if event.key == K_r:
                self.r = (self.r + 10) % 256
                self.text.setColor((self.r, self.b, self.g, 255))
            elif event.key == K_g:
                self.g = (self.g + 10) % 256
                self.text.setColor((self.r, self.b, self.g, 255))
            elif event.key == K_b:
                self.b = (self.b + 10) % 256
                self.text.setColor((self.r, self.b, self.g, 255))
            elif event.key == K_DOWN:
                self.posy += 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_UP:
                self.posy -= 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_RIGHT:
                self.posx += 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_LEFT:
                self.posx -= 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_y:
                self.rotation += 0.1
                self.text.setRotation(self.rotation)
            elif event.key == K_i:
                self.scaley -= 0.1
                self.text.setScaleY(self.scaley)
            elif event.key == K_k:  
                self.scaley += 0.1
                self.text.setScaleY(self.scaley)
            elif event.key == K_j:
                self.scalex -= 0.1
                self.text.setScaleX(self.scalex)
            elif event.key == K_l:
                self.scalex += 0.1
                self.text.setScaleX(self.scalex)
            elif event.key == K_y:
                self.rotation += 0.1
                self.text.setRotation(self.rotation)
                
    def loop(self, dt):
        self.z += dt
        if self.z > 500:
            self.z -= 500
            self.countedTime += 0.5
            caption = 'torture test pygame (instant) fps ' + \
                      str(round(self.clock.get_fps(), 2)) + ' averaged fps '\
                      + str(round((self.frames / self.countedTime), 2)) 
            pygame.display.set_caption(caption)
        self.frames += 1

        
if __name__ == '__main__':

    demo = TestApp()
    demo.run()
