"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

fbo.py

Provides access to FBOs.
"""

ENABLED = True

from collections import defaultdict
import weakref
from OpenGL.GL import glGetInteger, glDrawBuffers, GL_TEXTURE_2D, GL_NONE

from pygl3display.glcommoninst import GLCommon
try:
    from OpenGL.GL import (glGenFramebuffers, glFramebufferTexture2D, 
    glCheckFramebufferStatus, glBindFramebuffer, 
    GL_FRAMEBUFFER, GL_FRAMEBUFFER_COMPLETE,)
    import OpenGL.GL as _FBOLib
except ImportError:
    try:
        from OpenGL.GL.EXT.framebuffer_object import (
            glGenFramebuffersEXT as glGenFramebuffers, 
            glBindFramebufferEXT as glBindFramebuffer,
            glFramebufferTexture2DEXT as glFramebufferTexture2D,
            glCheckFramebufferStatusEXT as glCheckFramebufferStatus,
            GL_FRAMEBUFFER_EXT as GL_FRAMEBUFFER,
            GL_FRAMEBUFFER_COMPLETE_EXT as GL_FRAMEBUFFER_COMPLETE,
            GL_MAX_COLOR_ATTACHMENTS_EXT as GL_MAX_COLOR_ATTACHMENTS)
        import OpenGL.GL.EXT.framebuffer_object as _FBOLib
    except ImportError:
        ENABLED = False
        

class FBO(object):
    """An abstraction of an FBO"""
    def __init__(self):
        """Initialise the FBO"""
        self.framebuffer = glGenFramebuffers(1)
        self.textureSets = defaultdict(weakref.WeakValueDictionary)
        self.textureSetsBindLocs = defaultdict(dict)
        self.textureSetsFragLocs = defaultdict(dict)
        self.nextID = 0
    
    def getID(self):
        """Whilst a texture set can be any hashable object, getID will return
        something unique"""
        ret = self.nextID
        self.nextID += 1
        return ret
        
    def checkErrors(self):
        """Check to see if the FBO has any errors; if so, return error code,
        otherwise return True"""
        FBOCommonInst.bindFramebuffer(self)
        fbostatus = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        return True if fbostatus == GL_FRAMEBUFFER_COMPLETE else fbostatus
        
    def bind(self):
        """Bind the framebuffer"""
        FBOCommonInst.bindFramebuffer(self)
    
    @staticmethod
    def unbind():
        """Unbinds framebuffers"""
        FBOCommonInst.bindFramebuffer(None)
        
    def unbindTextureSet(self, textureSet):
        """Unbind all textures in the given texture set ID"""
        # Need a lazy call to glDrawTextures somewhere?
        # Only if necessary by further experiment
        textureSetDict = self.textureSets[textureSet]
        FBOCommonInst.bindFramebuffer(self)
        for location in textureSetDict:
            glFramebufferTexture2D(GL_FRAMEBUFFER, location, 
                GL_TEXTURE_2D, 0, 0)
        
    def bindTextureSet(self, textureSet):
        """Bind all textures in the given texture set ID"""
        textureSetDict = self.textureSets[textureSet]
        FBOCommonInst.bindFramebuffer(self)
        for location in textureSetDict:
            glFramebufferTexture2D(GL_FRAMEBUFFER, location,
                GL_TEXTURE_2D, textureSetDict[location].getTexture(), 0)
        bindDict = self.textureSetsFragLocs[textureSet]
        bindList = [bindDict[x] for x in xrange(len(bindDict))]
        FBOCommonInst.setBindPoints(bindList)
        
    def addToTextureSet(self, fraglocation, bindPoint, texture, textureSet):
        """Add a texture to a texture set
        Args:
            fraglocation: The location in the fragment array to bind to
            bindPoint: The attachment point to bind to
            texture: The texture
            textureSet: The ID of the texture set to add to"""
        bindPoint = FBOCommonInst.bindPoints.get(bindPoint, bindPoint)
        if bindPoint in self.textureSetsBindLocs[textureSet]:
            unboundFragLoc = self.textureSetsBindLocs[textureSet][bindPoint]
            #Delete stuff as necessary
            del self.textureSetsFragLocs[textureSet][unboundFragLoc]
        elif fraglocation in self.textureSetsFragLocs[textureSet]:
            unboundBindPoint = self.textureSetsFragLocs[textureSet][fraglocation]
            #Delete stuff as necessary
            del self.textureSetsBindLocs[textureSet][unboundBindPoint]
    
        self.textureSets[textureSet][bindPoint] = texture
        self.textureSetsFragLocs[textureSet][fraglocation] = bindPoint
        self.textureSetsBindLocs[textureSet][bindPoint] = fraglocation
        
            
    def bindTexture(self, location, texture):
        """Binds a texture onto the FBO"""
        location = FBOCommonInst.bindPoints.get(location, location)
        FBOCommonInst.bindFramebuffer(self)
        glFramebufferTexture2D(GL_FRAMEBUFFER, location, 
                               GL_TEXTURE_2D, texture.getTexture(), 0)
        
def importName(name):
    """Imports from the Framebuffer object library, taking into account
    that there may be an 'EXT' or '_EXT' appended."""
    names = [name, name + 'EXT', name + '_EXT']
    lookup = _FBOLib.__dict__
    for name in names:
        if name in lookup:
            return lookup[name]
    raise ImportError('FBO Bindings have no %s' % name)

if ENABLED:
    class FBOCommon(object):
        """Shared state between FBO objects"""
        def __init__(self):
            """Initialise the shared state"""
            self.boundFBO = 0
            self.fbos = defaultdict(FBO)
            self.boundBindPoints = []

        def __getattr__(self, key):
            """Enable lazy creation of the bindPoints dictionary"""
            if key in ('bindPoints', 'noColorBindPoints'):
                self.getBindPoints()
                return self.bindPoints
            else:
                return super(FBOCommon, self).__getattribute__(key)
                
        def setBindPoints(self, newBindPoints):
            """Set the currently bound points with GL_DRAW_BUFFERS"""
            if newBindPoints != self.boundBindPoints:
                if newBindPoints:
                    glDrawBuffers(newBindPoints)
                else:
                    glDrawBuffers([GL_NONE] * len(self.boundBindPoints))
                self.boundBindPoints = newBindPoints

        def getFBO(self, size):
            """Get the shared FBO for a given size; all textures in an
            FBO need to be the same size for FBOs to work on the OpenGL
            extension (as opposed to the OpenGL 3 core)"""
            return self.fbos[size]
                            
        def getBindPoints(self):
            """Construct the locations dictionary"""
            self.noColorBindPoints = glGetInteger(GL_MAX_COLOR_ATTACHMENTS)
            self.bindPoints = {}
            for x in xrange(self.noColorBindPoints):
                self.bindPoints['color%i' % x] = \
                    importName('GL_COLOR_ATTACHMENT%i'%x)
                
        def bindFramebuffer(self, fbo):
            """Lazily bind a framebuffer"""
            GLCommon.target = fbo
            fboval = 0 if fbo is None else fbo.framebuffer
            if self.boundFBO != fboval:
                glBindFramebuffer(GL_FRAMEBUFFER, fboval)
                self.boundFBO = fbo
    
    FBOCommonInst = FBOCommon()
    
    def unbindFBO():
        """Unbind any bound FBO; a hook for GLCommon, but needs to be in the
        FBO file as OpenGL 2.1 does not guarantee FBO's available"""
        FBOCommonInst.bindFramebuffer(None)
    GLCommon.addSwitchToMainTargetHook(unbindFBO)


        
