"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

rendertargets.py

Provides TextureTarget and TwoDTextureTarget, which allow 
render to texture support.
"""

from OpenGL.GL import glGetInteger
from pygl3display.twodtexture import TwoDTexture
from pygl3display.glcommoninst import GLCommon
from pygl3display.fbo import ENABLED

if not ENABLED:
    raise ImportError('FBOs not available')
    
from pygl3display.fbo import FBOCommonInst

__all__ = ['TextureTarget', 'TwoDTextureTarget']

class TextureTarget(object):
    """A Texture Target for render to multiple textures"""
    
    def __init__(self, number, size, compute = False):
        """Initialise a texture render target
        Args:
         number: Number of output textures
         size: Size of textures
         compute: If true, use GL_RGBA32 textures and disable blending
         for the target."""
        self.size = size
        self.compute = compute
        self.fbo = FBOCommonInst.getFBO(size)
        self.fboid = self.fbo.getID()
        self.compute = compute
        self.number = number
        self.textures = []
            
    def initTextures(self):
        """Initialise the textures of the texture target"""
        self.textures = []
        for x in xrange(self.number):
            texture = TwoDTexture(self.size, self.compute)
            self.textures.append(texture)
            self.fbo.addToTextureSet(x, 'color%i' % x, texture, self.fboid)
        self.fbo.bindTextureSet(self.fboid)
        status = self.fbo.checkErrors()
        if status is not True:
            raise Exception('FBO Status %i; not usable' % status)

    def __enter__(self):
        """Ensure that the render target is bound"""
        self.initTextures()
        self.bind()
        
    def __exit__(self, val, etype, traceback):
        """Remove the bindings of the render target"""
        self.unbind()
        
    def bind(self):
        """Bind the fbo and the texture set this render target uses"""
        self.fbo.bind()
        self.fbo.bindTextureSet(self.fboid)
        GLCommon.target = self
        GLCommon.setResolution((self.size, self.size))
        if self.compute:
            GLCommon.enableBlend(False)
        
    def unbind(self):
        """Unbind the texture set this render target uses"""
        self.fbo.unbindTextureSet(self.fboid)

        
class TwoDTextureTarget(TextureTarget):
    """A simple TwoD Texture Target that renders to a single target"""
    
    def __init__(self, size, compute = False):
        """A special case of a Texture Target, having only 1 texture"""
        super(TwoDTextureTarget, self).__init__(1, size, compute)
        self.texture = None 
        
    def __exit__(self, val, etype, traceback):
        """Remove the bindings of the render target"""
        super(TwoDTextureTarget, self).__exit__(val, etype, traceback)
        self.texture = self.textures[0]
        
    def readPixels(self):
        """Read the pixels out of the texture"""
        self.unbind()
        return self.texture.readPixels()
        
        
        
        
