"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

texture.py

Provides an abstract base for textures
"""

from pygl3display.uniforms import Uniform
import OpenGL.GL
import weakref
from OpenGL.GL import ( 
    glGenTextures, glBindTexture, glTexParameteri, glGetTexImage, glUniform1f, 
    glUniform1i, glActiveTexture, glDeleteTextures,
    GL_RGBA, GL_UNSIGNED_BYTE, GL_MAX_TEXTURE_COORDS, GL_TEXTURE_MIN_FILTER, 
    GL_TEXTURE_MAG_FILTER, GL_NEAREST, GL_FLOAT, GL_RED, GL_GREEN, GL_BLUE, 
    GL_ALPHA, GL_RGB, GL_BGR, GL_RGBA, GL_BGRA, GL_LUMINANCE, 
    GL_LUMINANCE_ALPHA)
    
__all__ = ['colorMasks', 'TextureCommonClass', 'Texture']
    
colorMasks = {'r': GL_RED, 'g': GL_GREEN, 'b': GL_BLUE, 'a': GL_ALPHA, 
    'rgb': GL_RGB, 'bgr': GL_BGR, 'rgba': GL_RGBA, 'bgra': GL_BGRA, 
    'l': GL_LUMINANCE, 'la': GL_LUMINANCE_ALPHA}

class TexUnitsDict(dict):
    """Maps integers to available texture units"""
    def __missing__(self, key):
        """Lazily create the map between integer i and GL_TEXTUREi"""
        if isinstance(key, int) and 0 <= key < GL_MAX_TEXTURE_COORDS:
            tuname = 'GL_TEXTURE%i' % key
            self[key] = OpenGL.GL.__dict__[tuname]
            return self[key]
        else:
            raise KeyError('%s is not a valid Texture Unit' % str(key))
        
class TextureCommonClass(object):
    """The shared state between all Textures"""
    
    def __init__(self):
        """Initialise the shared state"""
        self.texUnits = TexUnitsDict()
        self.usage = []
        self.activetu = 0
        
    def __getattr__(self, key):
        """Load the number of texture units on demand"""
        if key == 'numberOfTextureUnits':
            self.numberOfTextureUnits = GL_MAX_TEXTURE_COORDS
            # This should probably be moved to hacks after things have been
            # tracked down to specific drivers etc.
            if self.numberOfTextureUnits > 16: 
                self.numberOfTextureUnits = 16
            return self.numberOfTextureUnits
        else:
            super(self, TextureCommonClass).__getattribute__(key)
        
    def activateTU(self, tu):
        """Activates a texture unit. tu should be an 
        integer < GL_MAX_TEXTURE_COORDS."""
        if self.activetu != tu:
            glActiveTexture(self.texUnits[tu])
            self.activetu = tu
            
    def bindTextures(self, textures):
        """Checks to see if given textures could be bound simultaneously;
        binds them all if possible, otherwise raises an exception"""
        if len(textures) < self.numberOfTextureUnits:
            for texture in textures:
                self.bindTexture(texture)
        else:
            raise Exception('Only %i texture units available' 
                             % self.numberOfTextureUnits)
    
    def bindTexture(self, texture):
        """Binds a texture to a texture unit. Uses any available texture unit;
        informs the texture to update its tu uniform; Use an LRU cache to
        determine what should be removed"""
        for index in xrange(len(self.usage)):
            ref = self.usage[index][0]
            if ref() == texture:
                # Promote the texture to head of the queue and quit
                self.usage.append(self.usage.pop(index))
                return
            
        # Otherwise we've not found it and have to do something...
                
        if len(self.usage) >= self.numberOfTextureUnits:
            tu = self.usage.pop(0)[1]
        else:
            tu = len(self.usage)
        self.usage.append((weakref.ref(texture), tu))
        self.activateTU(tu)
        glBindTexture(texture.textureType, texture.surface)
        texture.setTU(tu)

TextureCommon = TextureCommonClass()

class Texture(object):
    """An abstract hardware accelerated texture"""
    
    def initTexture(self, size, textureType, initialiser):
        """Initialise the texture
        Args:
            size: Size of the texture
            textureType: 
            initiliser: a lambda function to initialise the texture
        """
        self.size = size
        self.textureType = textureType
        self.initialiser = initialiser
        
        self.tuUniform = Uniform(0, glUniform1i, [])
        self.tu = 0
        self.sizeUniform = Uniform(self.size, glUniform1f, [])
        self.texelUniform = Uniform(1.0 / self.size, glUniform1f, [])

        self.surface = glGenTextures(1)
        TextureCommon.bindTexture(self)
        self.initialiser()
        glTexParameteri(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        
    def __del__(self):
        """Free associated GL resources"""
        self.destroy()
        
    def setTU(self, tu):
        """Sets the texture unit the texture was last bound to"""
        self.tuUniform.uploadData(tu)
        self.tu = tu
                
    def getTexture(self):
        """Returns the gl Texture identifier"""
        return self.surface
        
    def destroy(self):
        """Destroy the gl texture"""
        if glDeleteTextures is not None: # May be none if program was killed
            glDeleteTextures(self.surface)
        self.surface = None
        
    def recreate(self):
        """Recreates the texture"""
        self.initTexture(self.size, self.textureType, self.initialiser)
        
    def bindTexture(self, layer=0):
        """Callback to bind the texture to the texture unit"""
        TextureCommon.bindTexture(self)
        
    def readPixels(self, mask = GL_RGBA, colorMasks = colorMasks):
        """Dump pixels from texture to a 3d numpy array. This is not
        really ideal, but alternatives are slow - this will be subject
        to change should I implement my own GL bindings.
        Arguments:
            mask: Either an OpenGL mask or an easier to use string
            from colorMasks"""
        mask = colorMasks.get(mask, mask)
        self.bindTexture()
        TextureCommon.activateTU(self.tu)
        nppixels = glGetTexImage(self.textureType, 0, mask, GL_FLOAT)
        #pixels = nppixels.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        return nppixels
        
    def readPixelsUB(self, mask = GL_RGBA, colorMasks = colorMasks):
        """Dump pixels as unsigned bytes. This returs a string, and is
        likely faster than anything readPixels will do. The returned string
        is also appropriate for uploading somewhere else. Of course,
        accuracy is lost."""
        mask = colorMasks.get(mask, mask)
        self.bindTexture()
        TextureCommon.activateTU(self.tu)
        pixels = glGetTexImage(self.textureType, 0, mask, GL_UNSIGNED_BYTE)
        return pixels
        

