"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

uniforms.py

Provides the representation of Uniforms
"""

import weakref
from pygl3display.glcommoninst import GLCommon

__all__ = ['Uniform', 'ProxyUniform', 'UniformStore']
    
class Uniform(object):
    """Represents a Uniform"""
    def __init__(self, data, func, args, defaultName = None):
        """Initialise the Uniform"""
        self.data = data
        self.func = func
        self.args = args
        self.names = weakref.WeakKeyDictionary()
        self.defaultName = defaultName
        self.shaderBindArgs = weakref.WeakKeyDictionary()
    
    def assignName(self, name, shader):
        """Informs the uniform of a name to use on a given shader"""
        self.names[shader] = name
    
    def dirty(self):
        """Mark the uniform as dirty"""
        for shader in self.shaderBindArgs:
            self.shaderBindArgs[shader][0] = True
            
    def uploadData(self, newData):
        """Upload new data to the uniform"""
        self.data = newData
        self.dirty()
        
    def genBindArgs(self, shader):
        """Generate arguments for binding the uniform on a shader"""
        
        if shader in self.names or self.defaultName is not None:
            name = self.names.get(shader, self.defaultName)
            location = shader.uniformLocs[name]
            self.shaderBindArgs[shader] = [True, [location] + self.args]
        else:
            raise Exception('Uniform does not have a name for given shader')
        
    def bind(self, shader, force):
        """Bind the uniform"""
        if shader not in self.shaderBindArgs:
            self.genBindArgs(shader)
        shaderBindArgs = self.shaderBindArgs[shader]
        if shaderBindArgs[0] or force:
            GLCommon.bindUniform(self.func, shaderBindArgs[1] + [self.data], 
                                force=shaderBindArgs[0])
            shaderBindArgs[0] = False
            
class ProxyUniform(object):
    """A proxy of a uniform"""
    def __init__(self):
        """Initialise variables"""
        self.__uniform = None
        
    def __getattr__(self, key):
        """Return the item from the proxied uniform"""
        return getattr(self.__uniform, key)
        
    def proxy(self, uniform):
        """Set the uniform to proxy"""
        if self.__uniform != uniform:
            self.__uniform = uniform
            uniform.dirty()
        
class UniformStore(object):
    """A UniformStore handles the storing and binding of Uniforms for a 
    shader"""
    def __init__(self, uniforms=None):
        """Initalise the uniform store.
        Args: 
          uniforms: A set of Uniform objects
        """
        self.uniforms = uniforms if uniforms is not None else set()
        self.shaderBindArgs = weakref.WeakKeyDictionary()
             
    def bind(self):
        """Bind the uniforms in this store"""
        shader = GLCommon.boundShader
            
        boundUniformStore = GLCommon.boundUniformStore
        if (shader not in boundUniformStore or 
           boundUniformStore[shader] is not self):
            boundUniformStore[shader] = self
            force = True
        else:
            force = False
        for uniform in self.uniforms:
            uniform.bind(shader, force)
                    
    def update(self, uniforms):
        """Update uniforms set"""
        self.uniforms.update(uniforms)
